"""
HELMpy, open source package of power flow solvers developed on Python 3 
Copyright (C) 2019 Tulio Molina tuliojose8@gmail.com and Juan José Ortega juanjoseop10@gmail.com

This program is free software: you can redistribute it and/or modify it under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
"""


# Import libraries and functions
import os
import pandas as pd
import numpy as np
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import spsolve
import cmath as cm
from time import time
import warnings
warnings.filterwarnings("ignore")
pd.set_option('display.max_rows',1000)
pd.set_option('display.max_columns',1000)
pd.set_option('display.width',1000)


#Global Variables that will be the basic parameters of the main function
detailed_run_print = False   #Print details on the run
Mis = 1e-4      #Mismatch
case = ''       #results file name
scale = 1       #load scale
iterations_limit = 15  #Max number of iterations per convergence check
Q_limits = True     #Checks Q generation limits

#Global variables declaration
N = 0; ref = 0; Y = 0;  Buses = [];
Yshunt = []; Ytrans = []; V = []; tita = []; tita_degree = []; iterations = 0; 
Pg = np.zeros(N); Pesp = np.copy(Pg); Qg = []; Pd = []; Qgmax = [];
Qgmin = []; Qd = []; Shunt = []; dimension = 0;
Pg_total = 0;  Qg_total = 0;  Pd_total = 0;
deltas_P_Q = 0; Jaco = 0; known = []; unknown = []; PVLIM_flag = 0; PVLIM_buses = False
deltas_tita_V = []; Plin_totales = 0; Qlin_totales = 0; Slin_total = 0
Qg_total_2 = 0; Ilineas = []; divergence = False
buses = 0
branches = 0
N_branches = 0
N_generators = 0
generators = 0
Number_bus = dict()
Ytrans_unsy = np.zeros((N,N), dtype=complex)
jaco_funct_store = []
Gen_contribute = []
Vre  = np.ones(N, dtype=float)
Vimag = np.zeros(N, dtype=float)
V_complex_profile = np.ones(N, dtype=complex)
Yre = np.zeros((N,N), dtype=float)
Yimag = np.zeros((N,N), dtype=float)
Pi = np.zeros(N, dtype=float)
Qi = np.zeros(N, dtype=float)
branches_buses = []
T_bucle_out = 0
known_dict = dict() 
unknown_dict = dict()
list_iterations = []
Power_branches = np.zeros((N_branches,8), dtype=float)
Ybr_list = list()
Power_print = pd.DataFrame()
Pmismatch = 0
S_gen = 0
S_load = 0
S_mismatch = 0
list_gen = np.zeros(1, dtype=int)

def variables_initialization():
    global N, ref, Buses, V, tita, Pg, Pesp, Qg, Pd, Qd, Qgmax, Qgmin, Shunt, dimension, Y
    global deltas_P_Q, Jaco, known, unknown, PVLIM_flag, PVLIM_buses, iterations
    global Yshunt, Ytrans, Number_bus
    global Ytrans_unsy, jaco_funct_store, Gen_contribute, Vre, Vimag, V_complex_profile
    global Yre, Yimag, Pi, Qi, branches_buses, list_iterations, Power_branches, Ybr_list, list_gen

    dimension = 0
    deltas_P_Q = 0
    Jaco = 0
    known = []
    unknown = []
    PVLIM_flag = 0
    iterations = 0
    list_iterations = []
    ref = 0
    PVLIM_buses = False
    Buses = [0 for i in range(N)]
    V = np.ones(N)
    tita = np.zeros(N)
    Pg = np.zeros(N)
    Qg = np.zeros(N)
    Pd = np.zeros(N)
    Qd = np.zeros(N)
    Qgmax = np.zeros(N)
    Qgmin = np.zeros(N)
    Y = np.zeros((N,N), dtype=complex)
    Shunt = np.zeros(N, dtype=complex)
    Yshunt = np.zeros((N,N), dtype=complex)
    Ytrans = np.zeros((N,N), dtype=complex)
    Pesp = np.copy(Pg)
    Number_bus = dict()
    Ytrans_unsy = np.zeros((N,N), dtype=complex)
    jaco_funct_store = []
    Gen_contribute = []
    Vre  = np.ones(N, dtype=float)
    Vimag = np.zeros(N, dtype=float)
    V_complex_profile = np.ones(N, dtype=complex)
    Yre = np.zeros((N,N), dtype=float)
    Yimag = np.zeros((N,N), dtype=float)
    Pi = np.zeros(N, dtype=float)
    Qi = np.zeros(N, dtype=float)
    branches_buses = [[i] for i in range(N)]
    Ybr_list = list()
    Power_branches = np.zeros((N_branches,8), dtype=float)
    list_gen = np.zeros(N_generators-1, dtype=int)

# Branches data processing to construct Ytrans, Yshunt, branches_buses and others
def Branches_processor(i, FromBus, ToBus, R, X, BTotal, Tap, Shift_degree):
    global Ytrans, Yshunt, Number_bus, branches_buses, Ybr_list, Ytrans_unsy
    FB = Number_bus[FromBus] 
    TB = Number_bus[ToBus]
    Ybr_list.append([FB, TB, np.zeros((2,2),dtype=complex)])
    Z = R + 1j*X
    if(Tap==0 or Tap==1):
        if(Z!=0):
            Yseries_ft = 1/Z
            if(Shift_degree==0):
                Ybr_list[i][2][0,1] = Ybr_list[i][2][1,0] = -Yseries_ft
            else:
                Shift = np.deg2rad(Shift_degree)
                Yseries_ft_shift = Yseries_ft/(np.exp(-1j*Shift))
                Yseries_tf_shift = Yseries_ft/(np.exp(1j*Shift))
                Ybr_list[i][2][0,1] = -Yseries_ft_shift
                Ybr_list[i][2][1,0] = -Yseries_tf_shift
                Ytrans_unsy[FB][TB] += Yseries_ft - Yseries_ft_shift
                Ytrans_unsy[TB][FB] += Yseries_ft - Yseries_tf_shift
            Ytrans[FB][TB] += -Yseries_ft
            Ytrans[FB][FB] +=  Yseries_ft
            Ytrans[TB][FB] += -Yseries_ft
            Ytrans[TB][TB] +=  Yseries_ft
        else:
            Ybr_list[i][2][0,1] = Ybr_list[i][2][1,0] = Yseries_ft = 0

        Bshunt_ft = 1j*BTotal/2
        Ybr_list[i][2][0,0] = Ybr_list[i][2][1,1] = Bshunt_ft + Yseries_ft
        Yshunt[FB][FB] +=  Bshunt_ft
        Yshunt[TB][TB] +=  Bshunt_ft
    else:
        Tap_inv = 1/Tap
        if(Z!=0):
            Yseries_no_tap = 1/Z
            Yseries_ft = Yseries_no_tap * Tap_inv
            if(Shift_degree==0):
                Ybr_list[i][2][0,1] = Ybr_list[i][2][1,0] = -Yseries_ft
            else:
                Shift = np.deg2rad(Shift_degree)                
                Yseries_ft_shift = Yseries_ft/(np.exp(-1j*Shift))
                Yseries_tf_shift = Yseries_ft/(np.exp(1j*Shift))
                Ybr_list[i][2][0,1] = -Yseries_ft_shift
                Ybr_list[i][2][1,0] = -Yseries_tf_shift
                Ytrans_unsy[FB][TB] += Yseries_ft - Yseries_ft_shift
                Ytrans_unsy[TB][FB] += Yseries_ft - Yseries_tf_shift
            Ytrans[FB][TB] += -Yseries_ft
            Ytrans[FB][FB] +=  Yseries_ft
            Ytrans[TB][FB] += -Yseries_ft
            Ytrans[TB][TB] +=  Yseries_ft 
        else:
            Ybr_list[i][2][0,1] = Ybr_list[i][2][1,0] = Yseries_no_tap = Yseries_ft = 0
        
        B = 1j*BTotal/2
        Bshunt_f = (Yseries_no_tap + B)*(Tap_inv*Tap_inv) 
        Bshunt_t = Yseries_no_tap + B
        Ybr_list[i][2][0,0] = Bshunt_f
        Ybr_list[i][2][1,1] = Bshunt_t
        Yshunt[FB][FB] +=  Bshunt_f - Yseries_ft
        Yshunt[TB][TB] +=  Bshunt_t - Yseries_ft


    if( TB not in branches_buses[FB] ):
        branches_buses[FB].append(TB)
    if( FB not in branches_buses[TB] ):
        branches_buses[TB].append(FB)

# Processing of .xlsx file data
def Buses_xls():
    global Buses, V, Qgmax, Qgmin, Pd, Qd, Pg, Shunt, buses, branches, N_branches
    global ref, N, N_generators, generators, Number_bus, Yshunt, Ytrans
    global Y, Ytrans_unsy, xls_actual, PVLIM_buses, Pesp, PVLIM_flag
    global Yre, Yimag, V_complex_profile, Vre, Vimag, scale, branches_buses, list_gen
    
    Pd = buses[2]/100*scale
    Qd = buses[3]/100*scale
    Shunt = buses[5]*1j/100 + buses[4]/100

    for i in range(N):
        Number_bus[buses[0][i]] = i
        if(buses[1][i]!=3):
            Buses[i] = 'PQ'
        else:
            ref = buses[0][i]
            aux_ref = i
        Yshunt[i][i] =  Shunt[i]

    pos = 0
    for i in range(N_generators):
        bus_i = Number_bus[generators[0][i]]
        if(bus_i!=aux_ref):
            list_gen[pos] = bus_i
            pos += 1
            Buses[bus_i] = 'PVLIM'
        PVLIM_buses = 1

        PVLIM_flag += 1
        V[bus_i] = generators[5][i]
        Pg[bus_i] = generators[1][i]/100*scale
        Qgmax[bus_i] = generators[3][i]/100
        Qgmin[bus_i] = generators[4][i]/100
        V_complex_profile[bus_i] = cm.rect(V[bus_i],tita[bus_i])
        Vre[bus_i]   = np.real(V_complex_profile[bus_i])
        Vimag[bus_i] = np.imag(V_complex_profile[bus_i])
       
    Buses[Number_bus[ref]] = 'Reference'
    Pg[Number_bus[ref]] = 0
    Pesp = np.copy(Pg)

    for i in range(N_branches):
        Branches_processor(i, branches[0][i], branches[1][i], branches[2][i], branches[3][i], branches[4][i], branches[8][i], branches[9][i])

    for i in range(N):
        branches_buses[i].sort()


# main function
def nr(GridName, Print_Details=False, Mismatch=1e-4, Results_FileName='', Scale=1, MaxIterations=15, Enforce_Qlimits=True):
    global Jaco, deltas_P_Q, deltas_tita_V, tita_degree, T_bucle_out
    global buses, branches, generators, N, N_generators, N_branches
    global detailed_run_print, Mis, case, scale, divergence, iterations_limit, Q_limits, list_iterations, iterations, V_complex_profile
    if (type(GridName)is not str) or(type(Print_Details)is not bool) or(type(Mismatch)is not float) or(type(Results_FileName)is not str) or not( (type(Scale)is float) or(type(Scale)is int) ) or(type(MaxIterations) is not int) or(type(Enforce_Qlimits) is not bool):
        print("Erroneous argument type.")
        return
    xls_actual = GridName
    detailed_run_print = Print_Details
    Mis = Mismatch
    if(Results_FileName==''):
        case = GridName[0:-5]
    else:
        case = Results_FileName
    scale = Scale
    iterations_limit = MaxIterations
    Q_limits = Enforce_Qlimits
    
    buses = pd.read_excel(xls_actual, sheet_name='Buses', header=None)
    branches = pd.read_excel(xls_actual, sheet_name='Branches', header=None)
    generators = pd.read_excel(xls_actual, sheet_name='Generators', header=None)
    N = len(buses.index)
    N_generators = len(generators.index)
    N_branches = len(branches.index)

    variables_initialization()
    Buses_xls()


def print_matrx(matrix):
    for i in range(matrix.shape[0]):
        print("Row %d"%i)
        for j in range(matrix.shape[1]):
            print(matrix[i,j],end='  ')
        print('\n')

#File
xls11 = 'case11.xlsx'

nr(xls11, Mismatch=1e-5, Scale=1, Print_Details=True)

# Ytrans  -------> Y series symmetric
# Ytrans_unsy ---> Y series unsymmetric (only the unsymetric part)
# Yshunt --------> Y shunt
   
Y = Ytrans + Yshunt + Ytrans_unsy

print("Matrix Ytrans:")
print_matrx(Ytrans)

print("Matrix Yshunt:")
print_matrx(Yshunt)

print("Matrix Ytrans_unsy:")
print_matrx(Ytrans_unsy)

print("Matrix Y:")
print_matrx(Y)
